window.MathJax = {
    tex: {
      macros: {
        bm: ["\\boldsymbol{#1}", 1],
        prefix: "\\psi",
        complete: "\\phi",
        pot: "\\Phi",
        xx: "\\bm{x}",
        A: "\\mathcal{A}",
        eos: "\\textsf{eos}",
        yy: "\\bm{y}"
      }
    },
    options: {
      processHtmlClass: 'arithmatex'
    }
  };
